/*
 * Copyright © 2020 Fachhochschule für die Wirtschaft (FHDW) Hannover
 *
 * This file is part of gaming-core.
 *
 * Gaming-core is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * Gaming-core is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * gaming-core. If not, see <http://www.gnu.org/licenses/>.
 */
package de.fhdw.gaming.core.domain;

import java.util.Optional;

/**
 * Represents a player, i.e.some participant of the game that acts for itself
 * and aims at winning the game. It does not necessarily have to be a human.
 * <p>
 * Note that CRTP (Curiously Recurring Template Pattern) is used to retain the
 * subtype of the player. This allows clients to clone a player via
 * {@link #deepCopy()} while retaining its type at compile-time.
 *
 * @param <P> The type of allowed players.
 */
public interface Player<P extends Player<P>> extends Stateful {

    /**
     * Returns the name of this player.
     */
    String getName();

    /**
     * Returns the state of this player.
     */
    PlayerState getState();

    /**
     * Changes the state of this player. If the new state is
     * {@link PlayerState#PLAYING}, any outcome previously set by calling
     * {@link #setOutcome(double)} is removed.
     *
     * @param newState The new state.
     */
    void setState(PlayerState newState);

    /**
     * Returns the outcome of this player (if available). The outcome is unavailable
     * if the player is in the state {@link PlayerState#PLAYING}. Otherwise, if no
     * outcome has been set, the outcome shall be computed from the player's state by calling
     * {@link AbstractPlayer#mapStateToOutcome(PlayerState)}.
     */
    Optional<Double> getOutcome();

    /**
     * Sets the outcome of this player. Requires that this player's state is not
     * {@link PlayerState#PLAYING}.
     *
     * @param newOutcome The new outcome.
     * @throws IllegalStateException if this player is in the state {@link PlayerState#PLAYING}.
     */
    void setOutcome(double newOutcome);

    @Override
    P deepCopy();
}
