/*
 * Copyright © 2020 Fachhochschule für die Wirtschaft (FHDW) Hannover
 *
 * This file is part of gaming-core.
 *
 * Gaming-core is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * Gaming-core is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * gaming-core. If not, see <http://www.gnu.org/licenses/>.
 */
package de.fhdw.gaming.core.domain;

import java.util.Optional;

/**
 * Represents a strategy. A strategy is bound to a {@link Player player} and determines her next move.
 *
 * @param <P> The type of allowed players.
 * @param <S> The type of allowed states.
 * @param <M> The type of allowed moves.
 */
public interface Strategy<P extends Player<P>, S extends State<P, S>, M extends Move<P, S>> {

    /**
     * Computes the next move.
     *
     * @param gameId The ID of the game.
     * @param player The player this strategy is bound to.
     * @param state  The game state. This is a copy of the game state. The game strategy is allowed to modify it at
     *               will, e.g. for computing possible move sequences in advance. However, changes have no effects
     *               whatsoever with regard to the game state.
     * @return The next move to finally apply to the game state. If {@link Optional#empty()} is returned, the strategy
     *         did not succeed in finding an appropriate move for the player. This causes the player to resign
     *         immediately.
     *         <p>
     *         Note that it is also possible to throw a {@link GameException} instead of returning an empty
     *         {@link Optional} (see below). The difference is that returning an empty {@link Optional} conveys the
     *         meaning of "resigning", whereas throwing a {@link GameException} signalises an unexpected error in the
     *         strategy's algorithm.
     * @throws GameException        if the strategy did not succeed in finding an appropriate move for the player due to
     *                              some unexpected game state. This causes the player to lose the game immediately.
     * @throws InterruptedException if choosing the move has been interrupted.
     */
    Optional<M> computeNextMove(int gameId, P player, S state) throws GameException, InterruptedException;

    /**
     * Returns {@code true} iff this strategy requires user interaction. Such a strategy is ignored when collecting
     * candidates for a game competition.
     * <p>
     * The default implementation returns {@code false}-
     */
    default boolean isInteractive() {
        return false;
    }

    /**
     * Requests the strategy to abort the move search being currently in-progress. Please note that the call is
     * typically done by some other (non-gaming) thread (e.g. the GUI), so take measures to avoid race conditions.
     *
     * @param gameId The ID of the game.
     */
    default void abortRequested(final int gameId) {
        // do nothing
    }

    /**
     * Requests the strategy to reset its internal state due to a new game being started. Each strategy may decide on
     * its own which state shall be reset and which state shall be retained. The latter allows self learning algorithms.
     */
    default void reset() {
        // do nothing
    }
}
